// PV112 2017, test for lesson 1

#include <memory>
#include <iostream>

#define GLEW_STATIC
#include "GL/glew.h"
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include "GL/freeglut.h"

using namespace std;

const int vertices_count = 3;
const float my_positions[vertices_count * 2] = {
     0.0f,  0.5f,
    -0.5f, -0.5f,
     0.5f, -0.5f,
};

const unsigned char my_colors[vertices_count * 3] = {
    255,   0,   0,
      0, 255,   0,
      0,   0, 255,
};

int win_width = 480;
int win_height = 480;

GLuint positions_vbo;
GLuint colors_vbo;
GLuint my_geometry_vao;
GLuint program;

GLint position_location;
GLint color_location;

void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    if (key == 27)
        exit(0);
}

void init()
{
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

    int compile_status;
    const char *vs_source =
        "#version 330\nin vec2 p;in vec3 c;out vec3 co;"
        "void main(){co=c;gl_Position=vec4(p,.0,1.0);}";

    GLuint vs_shader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vs_shader, 1, &vs_source, nullptr);
    glCompileShader(vs_shader);
    glGetShaderiv(vs_shader, GL_COMPILE_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetShaderiv(vs_shader, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetShaderInfoLog(vs_shader, log_len, nullptr, log.get());
        cout << "Failed to compile vertex shader" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    const char *fs_source =
        "#version 330\nout vec4 final_color;in vec3 co;"
        "void main(){final_color=vec4(co, 1.0);}";
    GLuint fs_shader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fs_shader, 1, &fs_source, nullptr);
    glCompileShader(fs_shader);
    glGetShaderiv(fs_shader, GL_COMPILE_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetShaderiv(fs_shader, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetShaderInfoLog(fs_shader, log_len, nullptr, log.get());
        cout << "Failed to compile fragment shader" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    program = glCreateProgram();
    glAttachShader(program, vs_shader);
    glAttachShader(program, fs_shader);
    glLinkProgram(program);
    glGetProgramiv(program, GL_LINK_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetProgramiv(program, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetProgramInfoLog(program, log_len, nullptr, log.get());
        cout << "Failed to link program" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    position_location = glGetAttribLocation(program, "p");
    color_location = glGetAttribLocation(program, "c");

    glGenBuffers(1, &positions_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, positions_vbo);
    glBufferData(GL_ARRAY_BUFFER, vertices_count * sizeof(float) * 2, my_positions, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glGenBuffers(1, &colors_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, colors_vbo);
    glBufferData(GL_ARRAY_BUFFER, vertices_count * sizeof(unsigned char) * 3, my_colors, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Create a vertex array object for the geometry
    glGenVertexArrays(1, &my_geometry_vao);

    // Set the parameters of the geometry
    glBindVertexArray(my_geometry_vao);
    glBindBuffer(GL_ARRAY_BUFFER, positions_vbo);
    glEnableVertexAttribArray(position_location);
    glVertexAttribPointer(position_location, 2, GL_FLOAT, GL_FALSE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, colors_vbo);
    glEnableVertexAttribArray(color_location);
    glVertexAttribPointer(color_location, 3, GL_UNSIGNED_BYTE, GL_TRUE, 0, 0);
    glBindVertexArray(0);
}

void render()
{
    glClear(GL_COLOR_BUFFER_BIT);

    glUseProgram(program);

    glBindVertexArray(my_geometry_vao);
    glDrawArrays(GL_TRIANGLES, 0, 3);

    glBindVertexArray(0);
    glUseProgram(0);

    glutSwapBuffers();
}

void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    glViewport(0, 0, win_width, win_height);
}

void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type)
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl;
        return;
    default:
        return;
    }
}

int main(int argc, char ** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS); // Return from the main loop

    glutInitContextVersion(3, 3);              // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);          // Use OpenGL debug context

    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - test1");

    glewExperimental = GL_TRUE;
    glewInit();

    PFNGLDEBUGMESSAGECALLBACKARBPROC myglDebugMessageCallbackARB;
#if defined(_WIN32) // Windows
    myglDebugMessageCallbackARB =
      (PFNGLDEBUGMESSAGECALLBACKARBPROC)wglGetProcAddress("glDebugMessageCallbackARB");
#elif defined(__APPLE__) // MacOS (not tested)
    myglDebugMessageCallbackARB = glDebugMessageCallbackARB;
#else // Linux
    myglDebugMessageCallbackARB =
      (PFNGLDEBUGMESSAGECALLBACKARBPROC)glXGetProcAddress((unsigned char *)"glDebugMessageCallbackARB");
#endif
    if (myglDebugMessageCallbackARB)
    {
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        myglDebugMessageCallbackARB(simple_debug_callback, nullptr);
    }

    init();

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);

    glutMainLoop();

    return 0;
}
